<?php

namespace DMG;

/**
 * Liveblog API
 * 
 * Register endpoints and filter reponses for liveblogs
 */
class Liveblog_API {

    /**
     * Bootstrap functionality
     *
     * @return void
     */
    public static function init(){
        add_action( 'rest_api_init', [ __CLASS__, 'register_routes' ] );
        add_filter( 'rest_prepare_post', [ __CLASS__, 'filter_post' ], 99, 3 );
        add_action( 'rest_prepare_' . Liveblog_Post_Type::NAME, [ __CLASS__, 'filter_liveblog_content' ], 10, 3 );
    }

    /**
     * Register the JSON API routes
     *
     * @return void
     */
    public static function register_routes() {

        /**
         * Route for getting a liveblog's entries
         *
         * Respects standard WP-API query parameters for pagination and ordering
         *
         * @example /wp-json/wp/v2/liveblog/10863242/entries
         *          Default entries request for Liveblog ID 10863242: first page, 10
         *          posts per page in descending order
         *
         * @example /wp-json/wp/v2/liveblog/10863242/entries?per_page=100&order=asc
         *          Returns the first 100 entries in ascending order
         *
         * @example /wp-json/wp/v2/liveblog/10863242/entries?page=5
         *          Returns the fifth page of results with default 10 entries per page
         */
        register_rest_route('wp/v2','/liveblog/(?P<parent>[\d]+)/entries',
            [
                'args'   => [
                    'parent' => [
                        'description' => __( 'The ID for liveblog parent.' ),
                        'type'        => 'integer',
                    ],
                ],
                [
                    'methods'             => \WP_REST_Server::READABLE,
                    'callback'            => [ __CLASS__, 'get_liveblog_entries' ],
                    'permission_callback' => '__return_true',
                ],
            ]
        );
    }

    /**
     * Return the liveblog entries for a given liveblog parent ID
     *
     * @param \WP_REST_Request $request
     * @return array The response
     */
    public static function get_liveblog_entries( \WP_REST_Request $request ) {

        $entries = [];

        // Get the liveblog ID
        $liveblog_id = $request->get_param( 'parent' );

        // Query parameters
        $per_page   = $request->get_param( 'per_page' );
        $page       = $request->get_param( 'page' );
        $order      = $request->get_param( 'order' );
        $after      = $request->get_param( 'after' );
        $before     = $request->get_param( 'before' );

        $query_args = [
                'post_type'      => Liveblog_Entry_Post_Type::NAME,
                'post_parent'    => (int) $liveblog_id,
                'posts_per_page' => ( is_numeric( $per_page ) ) ? (int) $per_page : 10,
                'paged'          => ( is_numeric( $page ) ) ? (int) $page : 1,
                'orderby'        => 'date',
                'order'          => ( is_string( $order ) && in_array( strtolower( $order ), [ 'asc', 'desc' ], true ) ) ? $order : 'desc'
        ];

        if ( ! empty( $after ) || ! empty( $before ) ) {

            if ( ! empty( $before ) ) {
                if ( is_numeric( $before ) ) {
                    $before = date( 'c', (int) $before );
                } else {
                    $before = str_replace( ' ', '+', $before );
                }
            } else {
                $before = date( 'c', time() );
            }

            if ( ! empty( $after ) ) {
                if ( is_numeric( $after ) ) {
                    $after = date( 'c', (int) $after );
                } else {
                    $after = str_replace( ' ', '+', $after );
                }
            } else {
                $after = date( 'c', 0 );
            }

            $query_args[ 'date_query' ] = [
                [
                    'after'     => $after,
                    'before'    => $before,
                    'inclusive' => true,
                ],
            ];
        }

        $query = new \WP_Query( $query_args );

        // Populate entries using core post controller
        if ( $query->have_posts() ) {
            $posts_controller = new \WP_REST_Posts_Controller( Liveblog_Entry_Post_Type::NAME );
            foreach ( $query->posts as $post ) {
                $item = $posts_controller->prepare_item_for_response( $post, $request );
                if ( ! empty( $item->data ) ) {
                    $entry = $item->data;
                    // Removing some properties that aren't useful
                    unset( $entry['guid'] );
                    unset( $entry['status'] );
                    unset( $entry['type'] );
                    $entry['_embedded'] = ( !empty( $entry[ '_embedded' ] ) ) ? $entry[ '_embedded' ] : [];

                    // get the id of the override author 
                    $author_override = (int) get_post_meta( $post->ID, 'dmg_liveblog_author_override', true );
                    
                    // Embed the author object unless it's be explictly hidden by a user
                    // If there is an author override, we embed that author and not the creator.
                    if ( false === (bool) get_post_meta( $post->ID, 'dmg_liveblog_hide_byline', true ) ){
                        $user_controller = new \WP_REST_Users_Controller( );
                        $author_byline_id = $author_override ? $author_override : $entry['author'];
                        $author = $user_controller->prepare_item_for_response( new \WP_User( $author_byline_id ), $request );
                        if ( !empty( $author ) && property_exists( $author, 'data' ) ) {
                            $embedded_author = $author->data;
                            if ( !empty( $embedded_author['yoast_head'] ) ) {
                                unset( $embedded_author['yoast_head'] );
                            }
                            $entry['_embedded']['author'][] = $embedded_author;
                        }
                    }

                    // Allow for custom filtering of the entry response
                    $entry = apply_filters( 'dmg_liveblog_entry_response', $entry, $post );
                    $entries[] = $entry;
                }
            }
        }

        // Create the response object
        $response = new \WP_REST_Response( $entries, 200);
        $response->header( 'X-WP-Total', $query->found_posts );
        $response->header( 'X-WP-TotalPages', $query->max_num_pages );

        return $response;
    }

    /**
     * Filter regular posts with embedded liveblog to include inline assets
     *
     * @param WP_REST_Response $response The response object
     * @param WP_Post          $post     The post object
     * @param WP_REST_Request  $request  The request object
     */
    public static function filter_post( $response, $post, $request ) {

        // Simple check for if post_content has a liveblog
        if ( is_a( $post, 'WP_Post' )
            && strpos( $post->post_content, '<!-- wp:dmg/liveblog' )
            && ! empty( $response->data['content']['rendered'] )
        ) {

            // Parse out blocks so we can determine the liveblog
            $blocks = parse_blocks( $post->post_content );

            foreach ( $blocks as $block ) {
                if ( isset( $block['blockName'] )
                    && 'dmg/liveblog' === $block['blockName']
                    && isset( $block['attrs']['id'] )
                ) {
                    // Append the inline assets
                    $response->data['content']['rendered'] .= \DMG\Liveblog_Post_Type::get_inline_assets( $block['attrs']['id'], $post->ID );

                    // Break early
                    break;
                }
            }
        }

        return $response;
    }

    /**
     * Filter the liveblog content
     *
     * @param WP_REST_Response $response The response object
     * @param WP_Post          $post     The post object
     * @param WP_REST_Request  $request  The request object
     */
    public static function filter_liveblog_content( $response, $post, $request ) {
        
        if ( isset( $response->data['content']['rendered'] ) ) {

            // Query latest 10 entries
            $query = new \WP_Query(
                [
                    'post_type'         => Liveblog_Entry_Post_Type::NAME,
                    'post_parent'       => $post->ID,
                    'posts_per_page'    => 3,
                    'orderby'           => 'date',
                    'order'             => 'desc',
                    'ignore_sticky'		=> true,
			        'no_found_rows'		=> true,
                ]
            );

            // Populate entries using core post controller
            $content = [];
            if ( $query->have_posts() ) {
                $posts_controller = new \WP_REST_Posts_Controller( Liveblog_Entry_Post_Type::NAME );
                foreach ( $query->posts as $entry ) {
                    $content[] = $entry->post_content;
                }
            }

            // Populate Key Events list
            $response->data['key_events'] = [];
            $key_events = new \WP_Query(
                [
                    'post_type'                 => Liveblog_Entry_Post_Type::NAME,
                    'post_parent'               => $post->ID,
                    'posts_per_page'            => 100,
                    'orderby'                   => 'date',
                    'order'                     => 'desc',
                    'ignore_sticky'             => true,
			        'no_found_rows'		        => true,
                    'update_post_term_cache'    => false,
			        'update_post_meta_cache'    => false,
                    'meta_query'                => [
                        [
                            'key'=> 'dmg_liveblog_key_event',
                            'value'=> '1',
                        ]
                    ]
                ]
            );

            if ( $key_events->have_posts() ) {
                foreach ( $key_events->posts as $key_event ) {
                    $key_event = $posts_controller->prepare_item_for_response( $key_event, $request );
                    $response->data['key_events'][] = $key_event->data;
                }
            }

            $response->data['schema'] = '<script type="application/ld+json">' . wp_json_encode( \DMG\Liveblog_Post_Type::generate_schema( $post->ID ) ) . '</script>';

            $response->data['content']['rendered'] = apply_filters( 'the_content', implode( "\n\n", $content ) );
        }

        return $response;
    }

}