/**
 * WordPress dependencies
 */
import { useState } from '@wordpress/element';
import { dispatch, select, withSelect } from '@wordpress/data';
import apiFetch from '@wordpress/api-fetch';

/**
 * Internal dependencies
 */
import { entriesPerPage } from "../../config";

/**
 * Load in entries
 */
export const loadEntries = (page) => {

    const postId = select('core/editor').getCurrentPostId();
    const endpoint = `wp/v2/liveblog/${postId}/entries?context=edit&per_page=${entriesPerPage}&page=${page}`

    // POST to create a new entry
    apiFetch({
        path: endpoint,
        method: 'GET',
        parse: false
    }).then((res) => {

        // Check if we've hit max pages
        const maxPages = res.headers.get('X-Wp-Totalpages');
        if (page >= maxPages) {
            dispatch('dmg/liveblog').setHasHitMaxPages(true);
        }

        return res.json();
    }).then((json) => {

        const rawBlocks = json.map((entry) => {
            return entry?.content?.raw;
        });

        // Filter out blocks that aren't already stored i.e. via Key Event selection
        const newBlocks = wp.blocks.parse(rawBlocks.join("\n\n")).filter((block) => {
            return (block?.attributes?.entryId && !select('dmg/liveblog').getBlockByEntryId(block.attributes.entryId));
        });

        // Add new blocks to the store
        newBlocks.forEach((block) => {
            if (block?.attributes?.entryId) {
                dispatch('dmg/liveblog').setBlockByEntryId(block.attributes.entryId, block);
            }
        });

        // Insert the new blocks
        if (!select('dmg/liveblog').hasFinishedInitialLoad()) {
            dispatch('core/block-editor').insertBlocks(newBlocks, 1, '', false);
            dispatch('dmg/liveblog').setHasFinishedInitialLoad(true);
        } else {
            dispatch('core/block-editor').insertBlocks(newBlocks, select('core/block-editor').getBlocks().length, '', false);
        }
    });
}

/**
 * Create a component for loading in more entries
 */
const LoadMore = ({hasHitMaxPages, hasFinishedInitialLoad}) => {

    const [page, setPage] = useState(1);

    // Load in more entries
    const loadMore = () => {
        const newPage = page + 1;
        setPage(newPage);
        loadEntries(newPage);
    }

    // CSS class names
    let cssClasses = ['dmg-liveblog__entry-load-more'];
    if (!hasFinishedInitialLoad) {
        cssClasses.push('dmg-liveblog__entry-load-more--loading');
    }

    // Button text display
    let text = (hasFinishedInitialLoad) ? 'Load More' : 'Loading entries...';

    return (
        <>
            {!hasHitMaxPages &&
                <div
                    className={cssClasses.join(' ')}
                    onClick={loadMore}>
                    {text}
                </div>
            }
        </>
    );
}

/**
 * Fetch the author
 *
 * @param {Object}   select                       Data store selector.
 * @return {Object} The liveblog object
 */
export default withSelect((select) => {

    // Get state flags
    const hasHitMaxPages = select('dmg/liveblog').hasHitMaxPages();
    const hasFinishedInitialLoad = select('dmg/liveblog').hasFinishedInitialLoad();

    return {
        hasHitMaxPages,
        hasFinishedInitialLoad
    };
})(LoadMore);
