
/**
 * External dependencies
 */
import React from 'react';
import { debounce } from 'lodash';
import moment from 'moment';

/**
 * WordPress dependencies
 */
import {
    CheckboxControl,
    Button,
    ComboboxControl,
    Spinner,
    Modal,
    TimePicker,
    PanelBody
} from '@wordpress/components';
import { 
    RichText,
    InnerBlocks,
    store as blockEditorStore,
    InspectorControls
} from "@wordpress/block-editor";
import { serialize } from "@wordpress/blocks";
import { select, dispatch, withSelect, useSelect } from '@wordpress/data';
import { useState, useEffect } from '@wordpress/element';
import apiFetch from '@wordpress/api-fetch';
import { innerBlocksAllowedBlocks } from "../../config";
import { store as coreStore } from '@wordpress/core-data';
import { date, getSettings } from '@wordpress/date';
import { buildQueryString } from "@wordpress/url";
import { close } from '@wordpress/icons';
/**
 * Internal dependencies
 */
import insertNewEntryForm from '../../utils'

// Define the REST API endpoint
const REST_ENDPOINT = '/wp/v2/dmg_liveblog_entry';

/**
 * Liveblog entry block
 *
 * @param {Object}   props                        The block props.
 * @param {Object}   props.attributes             Block attributes.
 * @param {Function} props.setAttributes          Sets the value for block attributes.
 * @param {string}   props.clientId               Client ID of the block.
 * @param {Object}   props.author                 The author object.
 * @param {Array}    props.keyEvents              Key event objects.
 * @return {Function} Render the edit screen
 */
const Block = ({ attributes, setAttributes, clientId, authorCreatedby, authorOverride, keyEvents }) => {

    const [isUpdating, setIsUpdating] = useState(false);
    const [isEditingAuthorOverride, setIsEditingAuthorOverride] = useState(false);
    const [authorDropdownOptions, setAuthorDropdownOptions ] = useState( [] );
    const [showingDeleteConfirm, setShowingDeleteConfirm] = useState(false);
    const [showDateOverride, setShowDateOverride] = useState(false);
    const [dateOverride, setDateOverride] = useState('');
    const [isSelected, setIsSelected] = useState(false);
    const [isFetching, setIsFetching] = useState(false);
    const [keyEventConfirm, setKeyEventConfirm] = useState( false );

    // Generate a stringified version a block for sync comparison,
    // removing whitespace and non-breaking spaces
    const getComparisonString = (rawBlock) => {
        return rawBlock
            .replace(/\n/g, '') // Newlines
            .replace(/&nbsp;/g,' ') // HTML non-breaking spaces
            .replace(/\xA0/g,' ') // Unicode non-breaking spaces
    }

    // the author object that we display in the frontend
    const author = authorOverride ? authorOverride : authorCreatedby;

    // Fetch the latest content
    const fetchLatestContent = () => {

        // Make sure entry ID is set so we can fetch
        if (attributes?.entryId) {

            // Start fetching
            setIsFetching(true);

            apiFetch({
                path: `${REST_ENDPOINT}/${attributes.entryId}?context=edit`,
                method: 'GET',
            }).then((res) => {

                // Sync the fetched block to the local block if they're not identical
                if (res?.content?.raw && getComparisonString(res.content.raw) !== getComparisonString(serialize(select('core/block-editor').getBlocksByClientId(clientId)[0]))) {
                    const parsedBlock = wp.blocks.parse(res.content.raw);
                    dispatch('core/block-editor').replaceBlock(clientId, parsedBlock);
                    dispatch('dmg/liveblog').setBlockByEntryId(attributes.entryId, parsedBlock[0]);
                }

                setIsFetching(false);
            }).catch((res) => {

                // Notify user that this post was deleted
                dispatch('core/notices').createNotice(
                    'info',
                    'This entry has been deleted by another user.',
                    {
                        id: 'dmg-liveblog-entry-deleted',
                        isDismissible: true,
                        type: 'snackbar',
                    }
                );

                // Remove the block
                dispatch('core/block-editor').removeBlock(clientId);
            });
        }
    }

    // Poll for new entries
    const pollNewEntries = () => {

        // Get 10 latest entry IDs locally
        const localTenIds = select('core/block-editor')
            .getBlocks()
            .map((block) => {
                return block.attributes.entryId
            })
            .filter((id) => {
                return id.length !== 0
            }).
            slice(0, 10);

        // Get the parent ID
        const parentId = select('core/editor').getCurrentPostId();

        // Fetch the latest 10 entries
        apiFetch({
            path: `wp/v2/liveblog/${parentId}/entries?context=edit`,
            method: 'GET',
        }).then((res) => {

            let keyOffset = 0;

            // Compare the fetched entries with the local IDs
            res.forEach( (entry, key) => {
                
                const keyToCheck = key - keyOffset;

                // Found a new entry; insert it into the editor
                if (entry.id !== localTenIds[keyToCheck]) {
                    const parsedBlock = wp.blocks.parse(entry.content.raw);
                    dispatch('core/block-editor').insertBlocks(parsedBlock, key + 1, '', false); // Add one to account for the blank entry form
                    keyOffset++;
                }

            });

        });

    }

    // Determine if there's been a change in block selection
    const { isBlockSelected, hasSelectedInnerBlock } = useSelect( blockEditorStore );
    if (!isSelected && (isBlockSelected(clientId) || hasSelectedInnerBlock(clientId, true))) {
        setIsSelected(true);
        fetchLatestContent();
    } else if(isSelected && !isBlockSelected(clientId) && !hasSelectedInnerBlock(clientId, true) ) {
        setIsSelected(false);
    }

    // Update the entry title
    const updateTitle = (newTitle) => {
        setAttributes({ title: newTitle });
    }

    // Update the key event bool
    const updateKeyEvent = (newBool) => {
        setAttributes({ keyEvent: newBool });
    }

    // Update the hide byline bool
    const updateHideByline = (newBool) => {
        setAttributes({ hideByline: newBool });
    }

    // Get timezone-adjusted date
    const getTimezoneAdjustedDate = (dateString) => {
        return moment.tz(dateString, getSettings().timezone.string).toISOString();
    }

    // Get data necessary to create and update an entry
    const getEntryData = () => {

        let entryData = {
            title: attributes.title,
            content: serialize(select('core/block-editor').getBlocksByClientId(clientId)[0]),
            parent: select('core/editor').getCurrentPostId(),
            dmg_liveblog_key_event: attributes.keyEvent,
            dmg_liveblog_hide_byline: attributes.hideByline,
            dmg_liveblog_author_override: attributes.authorOverride,
            status: 'publish'
        };

        // Only specify a date override when it's explicitly set
        if (showDateOverride && dateOverride) {
            entryData.date = getTimezoneAdjustedDate(dateOverride);
        }

        return entryData;
    }

    // Publish the entry
    const publishEntry = () => {

        setIsUpdating(true);

        // POST to create a new entry
        apiFetch({
            path: REST_ENDPOINT,
            method: 'POST',
            data: getEntryData(),
        }).then((res) => {

            // Reset the date override
            if (showDateOverride) {
                setShowDateOverride(false);
                setDateOverride('');
            }

            // Returned an ID; update the entryID attribute
            if (res?.id) {
                setAttributes({ entryId: res.id });
            }

            // Returned a date; update the date attribute
            if (res?.date) {
                setAttributes({ date: getTimezoneAdjustedDate(res.date) });
            }

            // Returned an author; update the author attribute
            if (res?.author) {
                setAttributes({ author: res.author });
            }

            // Update the list of key events, if applicable
            if (res?.dmg_liveblog_key_event) {
                dispatch('dmg/liveblog').addKeyEvent(res);
            }

            pollNewEntries();

            setIsUpdating(false);

            // Follow-up request to sync block string back to post object
            if (res?.id) {
                apiFetch({
                    path: `${REST_ENDPOINT}/${res.id}`,
                    method: 'POST',
                    data: getEntryData(),
                })
            }

            // Insert a new entry form
            insertNewEntryForm();
        });
    }

    // Update an entry
    const updateEntry = () => {

        setIsUpdating(true);

        // POST to create a update the entry
        apiFetch({
            path: `${REST_ENDPOINT}/${attributes.entryId}`,
            method: 'POST',
            data: getEntryData(),
        }).then((res) => {

            // We overrode the date; replace the block attribute and reset our states
            if (showDateOverride && res?.date) {
                setAttributes({ date: getTimezoneAdjustedDate(res.date) });
                setShowDateOverride(false);
                setDateOverride('');

                // Sync block attributes back
                apiFetch({
                    path: `${REST_ENDPOINT}/${attributes.entryId}`,
                    method: 'POST',
                    data: getEntryData(),
                })
            }

            // Update the list of key events, if applicable
            if (res?.dmg_liveblog_key_event) {
                dispatch('dmg/liveblog').addKeyEvent(res);
            } else {
                dispatch('dmg/liveblog').removeKeyEventByEntryId(res.id);
            }

            pollNewEntries();

            setIsUpdating(false);
        });
    }

    // Delete an entry
    const deleteEntry = () => {

        setIsUpdating(true);

        const entryIdToDelete = attributes.entryId;

        // POST to create a update the entry
        apiFetch({
            path: `${REST_ENDPOINT}/${entryIdToDelete}?force=true`,
            method: 'DELETE',
        }).then((res) => {
            dispatch('core/block-editor').removeBlock(clientId);
            dispatch('dmg/liveblog').removeKeyEventByEntryId(entryIdToDelete);
            pollNewEntries();
            setIsUpdating(false);
        });
    }

    // Output the date object
    const outputDate = (entryDate) => {
        const datetimeFormat = dmgLiveblogSettings?.datetimeFormat ?? getSettings().datetime;
        return date(dmgLiveblogSettings.datetimeFormat, new Date(entryDate));
    }

    // better UX when moving to the edition of the author byline: 
    // the cursor sets into the input lookup automatically,
    // and we apply the behaviour of cancelling the edition when clickingg on the X
    useEffect( ()=> { 
        if (isEditingAuthorOverride === 'editing') {
            document.querySelector(`#author-override-lookup-${clientId} input`)?.focus();
            document.querySelector(`#author-override-lookup-${clientId} .components-combobox-control__reset svg`)?.addEventListener( 'click', () => setIsEditingAuthorOverride(false) )
        }
    }, [isEditingAuthorOverride] );

    // triigered onkey press inside the lookup filter input for author.
    const handleOnFilterUsers = React.useMemo(
        () => debounce((inputText) => {
                // make api call to authors containing the text inputText
                const users = apiFetch( { path: `/wp/v2/users/?${buildQueryString({'per_page': 3, 'search': inputText })}` } )
                    .then( (users) => {
                        if (Array.isArray(users)) {
                            // sometimes the users API return users whose occurrence with the inputText is only
                            // in the email, but not in the name. That's why it won't show any result even if there are some.
                            // We need to apply this modification in order to show the results.
                            const hasUsersWithNameContainingInput = users.some(u => u.name.includes(inputText));
                            users.forEach( u => {
                                if (!u.name.includes(inputText)) {
                                    u.name += ` (email contains text '${inputText}')`;
                                }
                            });
                            
                            if (users.length) {
                                const authorDropdown = users.map( user => { return { 'value': user.id , 'label': user.name  } } );
                                setAuthorDropdownOptions(authorDropdown);
                            } else {
                                setAuthorDropdownOptions([{'value':null, 'label': `No results for '${inputText}'`}]);
                            }

                            if (!hasUsersWithNameContainingInput) {
                                setIsEditingAuthorOverride('hint:write a longer text to find the user');
                            } else if (users.length === 1) {
                                setIsEditingAuthorOverride('hint:click enter to select the user');
                            } else {
                                setIsEditingAuthorOverride('editing');
                            }
                        };
                    });
        }, 400),
        []
    );


    // CSS class names
    let cssClasses = ['dmg-liveblog__entry'];
    if (attributes.hideByline) {
        cssClasses.push('dmg-liveblog__entry--hide-byline');
    }

    if (isFetching) {
        cssClasses.push('dmg-liveblog__entry--is-fetching')
    }

    // Define the currentTime attribute for our TimePicker
    // Need to use en-US to ensure <TimePicker /> gets the
    // correct date format
    const currentDate = new Date(attributes?.date ?? null).toLocaleString('en-US', {
        timeZone: getSettings().timezone.string,
    });

    // Key event output
    const keyEventPosts = (typeof(keyEvents) === 'object') ? keyEvents : [];
    const outputKeyEvents = () => keyEventPosts.map((keyEvent, index) => {

        // Parse the block so we can output the title and date
        const parsedBlock = wp.blocks.parse(keyEvent?.content?.raw);
        let eventDate = '';
        if(parsedBlock.length === 1){
            eventDate = parsedBlock[0].attributes?.date;
        }
        return (
            <li key={`key-event-${index}`}
                className="dmg-liveblog__key-event"
                onClick={(e) => setActiveKeyEvent(e, keyEvent)}>
                <div>
                    <h4
                        dangerouslySetInnerHTML={{
                          __html: keyEvent.title.rendered
                        }}>
                    </h4>
                    <span
                        className="dmg-liveblog__key-event-date">
                        {outputDate(eventDate)}
                    </span>
                </div>
                <div>
                    <span
                        className="dashicons dashicons-star-filled"
                        onMouseOver={({ target }) => (target.className = "dashicons dashicons-star-empty")}
                        onMouseOut={({ target }) => (target.className = "dashicons dashicons-star-filled")}
                        onClick={() => setKeyEventConfirm(keyEvent)}>
                    </span>
                </div>
            </li>
        );
    });

    // Set a key event as active
    const setActiveKeyEvent = (e, keyEvent) => {

        // Get the entry from the custom liveblog store
        const keyEventEntry = select('dmg/liveblog').getBlockByEntryId(keyEvent.id);

        if (keyEventEntry?.clientId) {

            // Found an entry; select the block
            dispatch('core/block-editor').selectBlock(keyEventEntry.clientId);
        } else {

            // Didn't find a block for the key event; query for it
            apiFetch({
                path: `${REST_ENDPOINT}/${keyEvent.id}?context=edit`,
                method: 'GET',
            }).then((res) => {

                // Append the entry to the end and select it after it's inserted
                if (res?.content?.raw && res.content.raw !== serialize(select('core/block-editor').getBlocksByClientId(clientId)[0])) {
                    const parsedBlock = wp.blocks.parse(res.content.raw);
                    if (parsedBlock.length === 1 && parsedBlock[0]?.clientId) {
                        dispatch('core/block-editor').insertBlocks(parsedBlock);
                        dispatch('core/block-editor').selectBlock(parsedBlock[0].clientId);
                        dispatch('dmg/liveblog').setBlockByEntryId(keyEvent.id, parsedBlock[0]);
                    }
                }
            })
        }

        // If this click was on the key event star, set the modal active
        if (e.target.classList.contains('dashicons-star-empty') || e.target.classList.contains('dashicons-star-filled')) {
            setKeyEventConfirm(keyEvent);
        }
    }

    // On confirmation of key event removal dialog
    const onKeyEventRemovalConfirm = (keyEventToRemove) => {

        // Update the block attributes
        const blockToUpdate = select('dmg/liveblog').getBlockByEntryId(keyEventToRemove.id);
        if (blockToUpdate?.clientId) {

            // Update the block attributes
            dispatch('core/block-editor').updateBlockAttributes(blockToUpdate.clientId, {
                keyEvent: false
            });

            // POST to create a update the entry
            apiFetch({
                path: `${REST_ENDPOINT}/${keyEventToRemove.id}`,
                method: 'POST',
                data: {
                    content: serialize(select('core/block-editor').getBlocksByClientId(blockToUpdate.clientId)),
                    dmg_liveblog_key_event: false
                },
            }).then((res) => {

                // Remove key event from store
                dispatch('dmg/liveblog').removeKeyEventByEntryId(keyEventToRemove.id);

                // Close the dialog
                setKeyEventConfirm(false);
            })
        }

    }

    const hasAMPM = dmgLiveblogSettings?.datetimeFormat && dmgLiveblogSettings.datetimeFormat.match(/G|H/) ? false : true

    return (
        <>
            <InspectorControls>
                <PanelBody
                    title="Key Events"
                    className="dmg-liveblog__key-event-selector"
                    initialOpen>
                    <ul className="dmg-liveblog__key-events striped">
                        {outputKeyEvents()}
                    </ul>
                </PanelBody>
            </InspectorControls>
            <div className={cssClasses.join(' ')}>
                <div className="dmg-liveblog_entry_marker">
                </div>
                <div className="dmg-liveblog__entry-content">
                    <div className="dmg-liveblog__entry-date">
                        {!showDateOverride && attributes.date &&
                            <span
                                onClick={() => setShowDateOverride((!showDateOverride))}
                            >
                                {outputDate(attributes.date)}
                            </span>
                        }
                        {showDateOverride &&
                            <TimePicker
                                currentTime={ currentDate }
                                onChange={(value) => setDateOverride(value)}
                                is12Hour={ hasAMPM }
                            />
                        }
                    </div>
                    <RichText
                        className="dmg-liveblog__entry-title"
                        value={attributes.title}
                        onChange={(value) => updateTitle(value)}
                        placeholder={'Enter an Entry Title'}
                        keepPlaceholderOnFocus
                    />
                    { isEditingAuthorOverride ? (
                        <div id={ `author-override-lookup-${clientId}` } style={{position:'relative'}} className='dmg-liveblog__entry-author-override'>
                            <span className="dmg-liveblog__entry-author-override--creator">Created by <b>{authorCreatedby?.name}</b> </span>
                            { isEditingAuthorOverride === 'loading' && <Spinner/> }
                            { isEditingAuthorOverride.includes('warning:') && <small style={{position:'absolute',right:0,color:'red'}}>{isEditingAuthorOverride.replace('warning:','')}</small> }
                            { isEditingAuthorOverride.includes('hint:') && <small style={{position:'absolute',right:0,color:'gray'}}>{isEditingAuthorOverride.replace('hint:','')}</small> }
                            <ComboboxControl
                                label={ 'Select An Author' }
                                options={ authorDropdownOptions !== false ? authorDropdownOptions : [] }
                                onChange={ ( newAuthorValue ) => {
                                    setAttributes( { authorOverride: newAuthorValue } );
                                    setIsEditingAuthorOverride(false);
                                } }
                                onFilterValueChange={ (inputText) => { 
                                    if (inputText.length < 3) {
                                        setIsEditingAuthorOverride('warning:Type at least 3 characters');
                                        setAuthorDropdownOptions([]);
                                    } else {
                                        setIsEditingAuthorOverride('loading');
                                        setAuthorDropdownOptions([{value:null, label:'Looking for users with text "'+inputText+'"'}]);
                                        handleOnFilterUsers(inputText); 
                                    }
                                } }
                            />
                        </div>
                        ) 
                        :
                        <div className={`dmg-liveblog__entry-author-fields ${ attributes.authorOverride? 'dmg-liveblog__entry-author-fields--overridden' : '' }` }>
                            { !author && <Spinner/> }
                            <div onClick={ () => setIsEditingAuthorOverride('editing') }>
                                {author?.name &&
                                    <span className="dmg-liveblog__entry-author">{author?.name}</span>
                                }
                            </div>
                            <CheckboxControl
                                    className="dmg-liveblog__hide-byline-checkbox"
                                    checked={attributes.hideByline}
                                    onChange={updateHideByline}
                                />
                            { attributes.authorOverride && <Button
                                    icon={ close }
                                    iconSize={ 20 }
                                    className={ 'dmg-liveblog__cancel-byline-overridden' }
                                    label={ 'Reset the author to the creator of the entry' }
                                    onClick={ () => setAttributes( {authorOverride: null} ) }
                                /> }
                        </div>
                    }
                    <InnerBlocks allowedBlocks={innerBlocksAllowedBlocks} />
                </div>
                <div className="dmg-liveblog__entry-toolbar">
                    <CheckboxControl
                        className="dmg-liveblog__key-event-checkbox"
                        label={<strong>Key event<span className="dashicons dashicons-star-filled"></span></strong>}
                        checked={attributes.keyEvent}
                        onChange={updateKeyEvent}
                    />
                    {!attributes.entryId &&
                        <div className="dmg-liveblog__entry-toolbar-actions">
                            {isUpdating &&
                                <Spinner />
                            }
                            <Button
                                variant="primary"
                                onClick={publishEntry}>
                                Publish
                            </Button>
                        </div>
                    }
                    {attributes.entryId &&
                        <div className="dmg-liveblog__entry-toolbar-actions">
                            {isUpdating &&
                                <Spinner />
                            }
                            <Button
                                variant="primary"
                                onClick={updateEntry}>
                                Update
                            </Button>
                            <Button
                                variant="primary"
                                onClick={() => setShowingDeleteConfirm(true)}
                                isDestructive>
                                Delete
                            </Button>
                        </div>
                    }
                </div>
            </div>
            {showingDeleteConfirm &&
                <Modal
                    title="Delete this entry?"
                    onRequestClose={() => setShowingDeleteConfirm(false)}
                    className="dmg-liveblog__modal">
                    <p>This will permanently remove this liveblog entry.</p>
                    <Button
                        variant="secondary"
                        onClick={() => setShowingDeleteConfirm(false)}>
                        Cancel
                    </Button>
                    <Button
                        variant="primary"
                        onClick={deleteEntry}>
                        Delete
                    </Button>
                </Modal>
            }
            {keyEventConfirm &&
                <Modal
                    title="Remove entry as key event?"
                    onRequestClose={() => setKeyEventConfirm(false)}
                    className="dmg-liveblog__modal">
                    <p>Are you sure you want to unmark <strong>{keyEventConfirm.title.rendered}</strong> as a key event?</p>
                    <Button
                        variant="secondary"
                        onClick={() => setKeyEventConfirm(false)}>
                        Cancel
                    </Button>
                    <Button
                        variant="primary"
                        onClick={() => onKeyEventRemovalConfirm(keyEventConfirm)}>
                        Remove
                    </Button>
                </Modal>
            }
        </>
    );
}

/**
 * Fetch the author and key events
 *
 * @param {Object}   select                       Data store selector.
 * @param {Object}   props                        The block props.
 * @param {Object}   props.attributes             Block attributes.
 * @return {Object} The liveblog author and key events
 */
export default withSelect((select, { attributes }) => {

    const { getEntityRecord } = select(coreStore);

    // Use getEntityRecord to fetch the author object
    const authorCreatedby = attributes?.author ? getEntityRecord('root', 'user', attributes.author) : null;
    const authorOverride  = attributes?.authorOverride ? getEntityRecord('root', 'user', attributes.authorOverride) : null;

    // Key events
    const keyEvents = select('dmg/liveblog').getKeyEvents();

    return {
        authorCreatedby,
        authorOverride,
        keyEvents
    };
})(Block);
