
/**
 * External dependencies
 */
import React, { Component } from 'react';

/**
 * WordPress dependencies
 */
import { PlainText } from '@wordpress/editor';
import { InnerBlocks, InspectorControls } from '@wordpress/block-editor';
import { PanelBody, Spinner } from '@wordpress/components';
import { withSelect } from '@wordpress/data';
import { store as coreStore } from '@wordpress/core-data';
import { RawHTML } from '@wordpress/element';

/**
 * Internal dependencies
 */
import LiveblogSelector from './liveblog-selector';


/**
 * Liveblog block
 *
 * @param {Object}   props                        The block props.
 * @param {Object}   props.attributes             Block attributes.
 * @param {Function} props.setAttributes          Sets the value for block attributes.
 * @return {Function} Render the edit screen
 */
const Block = ({ attributes, setAttributes, liveblog }) => {
    
    /**
     * Set attributes when a liveblog is selected
     */
    const onLiveblogSelect = ( liveblog ) => {
        setAttributes( {
            id: liveblog.id
        } );
    };

    return (
        <>
            <InspectorControls>
                <PanelBody title="Select a liveblog" initialOpen>
                    <LiveblogSelector
                        onChange={ onLiveblogSelect }
                        selected={ attributes?.id }
                    />
                </PanelBody>
            </InspectorControls>
            { liveblog === false && ! attributes?.id &&
                <div className="components-placeholder is-large">
                    <div className="components-placeholder__fieldset">
                        Please select a liveblog
                    </div>
                </div>
            }
            { typeof( liveblog ) !== 'object' && attributes?.id &&
                <div className="dmg-liveblog__loading">
                    <Spinner />
                    <span>Loading liveblog...</span>
                </div>
            }
            { typeof( liveblog ) === 'object' &&
                <div className="dmg-liveblog">
                    { liveblog?.title?.rendered &&
                        <>
                            <h3 className="dmg-liveblog__title">
                                <RawHTML>{ liveblog.title.rendered }</RawHTML>
                            </h3>
                            <div className="dmg-liveblog__entries">
                                <RawHTML>{ liveblog.content.rendered }</RawHTML>
                            </div>
                        </>
                    }
                </div>
            }
        </>
    );

}

/**
 * Fetch the liveblog
 *
 * @param {Object}   select                       Data store selector.
 * @param {Object}   props                        The block props.
 * @param {Object}   props.attributes             Block attributes.
 * @return {Object} The liveblog object
 */
export default withSelect( ( select, { attributes } ) => {

    const { getEntityRecord } = select( coreStore );

    // Use getEntityRecord to fetch liveblog
    let liveblog = false;
    if ( attributes?.id && attributes.id > 0 ) {
        liveblog = getEntityRecord( 'postType', 'dmg_liveblog', attributes.id, { context: 'view' } );
    }

    return {
        liveblog
    };
} )( Block );
