/**
 * WordPress dependencies
 */
import { withSelect } from '@wordpress/data';
import { TextControl, CheckboxControl, Spinner } from '@wordpress/components';
import { useState, RawHTML } from '@wordpress/element';
import { store as coreStore } from '@wordpress/core-data';

/**
 * External depedencies
 */
import { debounce } from 'lodash';

// Declare the number of search results per page
const RESULTS_PER_PAGE = 10;

/**
 * Liveblog selector component
 */
const LiveblogSelector = ( { liveblogs, setSearchTerm, page, setPage, onChange, selected } ) => {

	/**
	 * Generate list of liveblogs to display
	 */	
	const liveblogDisplay = ( liveblog ) => {

		let className = 'dmg-liveblog-selector__liveblog ';
		className += ( selected === liveblog.id ) ? 'dmg-liveblog-selector__liveblog-selected' : '';

		return (
			<li className={ className } onClick={ () => onChange( liveblog ) }>
				<div className="dmg-liveblog-selector__liveblog-title">
					<RawHTML>{ liveblog.title.rendered }</RawHTML>
				</div>
				<date>
					{ new Date( Date.parse( liveblog.date ) ).toDateString() }
				</date>
			</li>
		);
	};

	/**
	 * Debounced search function
	 */
	const search = debounce( (searchInput ) => {
		setSearchTerm( searchInput );
		setPage( 1 );
	}, 500 );

	/**
	 * Reset pagination on filter toggle
	 */
	const onFilterChange = (value) => {
		setPage( 1 );
	}

	// Determine if we need to hide any navigation
	const navPreviousClassname = ( page > 1 ) ? '' : 'hidden';
	const navNextClassname     = ( liveblogs?.length === RESULTS_PER_PAGE ) ? '' : 'hidden';

	return (
		<div className="dmg-liveblog-selector">
			<div className="dmg-liveblog-selector__search">
				<TextControl
					onChange={ search }
					placeholder="Enter a search term or liveblog ID"
				/>
			</div>
			{ liveblogs === null &&
				<>
					<Spinner />
					<span>
						<em>Loading...</em>
					</span>
				</>
			}
			{ liveblogs !== null && liveblogs.length === 0 &&
				<span>
					<em>No liveblogs found</em>
				</span>
			}
			{ liveblogs !== null && liveblogs.length > 0 &&
				<ul className="dmg-liveblog-selector__liveblog-list striped">
					{ liveblogs.map( liveblogDisplay ) }
				</ul>
			}
			<nav className="dmg-liveblog-selector__nav">
				<ul>
					<li
						className={ navPreviousClassname }
						onClick={ () => setPage( page - 1 ) }
					>
						&laquo; Previous
					</li>
					<li
						className={ navNextClassname }
						onClick={ () => setPage( page + 1 ) }
					>
						Next &raquo;
					</li>
				</ul>
			</nav>
		</div>
	);
};

export default withSelect( ( select, { postType, params } ) => {

	const { hasFinishedResolution } = select( coreStore );

	// Create some states
	const [ searchTerm, setSearchTerm ] = useState( '' );
	const [ page, setPage ] = useState( 1 );

	let liveblogs = [];

	// Check if this is a regular, alphanumeric search term
	if ( searchTerm === '' || isNaN( searchTerm ) ) {

		// Build array of query parameters
		const additionalParams = {
			page,
			context: 'view',
			per_page: RESULTS_PER_PAGE
		};

		if ( searchTerm !== '' ) {
			additionalParams.search = searchTerm;
		}

		const queryParams = { ...params, ...additionalParams };

		// Use getEntityRecords to fetch posts
		liveblogs = select( 'core' ).getEntityRecords( 'postType', 'dmg_liveblog', queryParams );
		if (!hasFinishedResolution('getEntityRecords', [ 'postType', 'dmg_liveblog', queryParams ])) {
			liveblogs = null;
		}
	} else {

		// Found a number, try to query by liveblog ID
		const liveblog = select('core').getEntityRecord( 'postType', 'dmg_liveblog', searchTerm );
		if ( liveblog !== null && liveblog?.id ) {
			liveblogs.push( liveblog );
		}
	}

	return {
		liveblogs,
		setSearchTerm,
		page,
		setPage
	};
} )( LiveblogSelector );
