/**
 * WordPress dependencies
 */
import { createReduxStore, register } from '@wordpress/data';

const DEFAULT_STATE = {
    blocks: {},
    hasFinishedInitialLoad: false,
    hasHitMaxPages: false,
    keyEvents: []
};

const actions = {

    // Add a block to the store by its post ID
    setBlockByEntryId( entryId, block ) {
        return {
            type: 'SET_BLOCK_BY_ENTRY_ID',
            entryId,
            block,
        };
    },

    // Set flag when we've finished initial load
    setHasFinishedInitialLoad( hasFinishedInitialLoad ) {
        return {
            type: 'SET_HAS_FINISHED_INITIAL_LOAD',
            hasFinishedInitialLoad,
        };
    },

    // Set flag when we've queried the max number of pages
    setHasHitMaxPages( hasHitMaxPages ) {
        return {
            type: 'SET_HAS_HIT_MAX_PAGES',
            hasHitMaxPages,
        };
    },

    // Set the entire array of key events
    setKeyEvents( keyEvents ) {
        return {
            type: 'SET_KEY_EVENTS',
            keyEvents
        }
    },

    // Add key event to store
    addKeyEvent( entry ) {
        return {
            type: 'ADD_KEY_EVENT',
            entry
        }
    },

    // Remove key event from store
    removeKeyEventByEntryId( entryId ) {
        return {
            type: 'REMOVE_KEY_EVENT_BY_ENTRY_ID',
            entryId
        }
    },
};

const store = createReduxStore( 'dmg/liveblog', {
    reducer( state = DEFAULT_STATE, action ) {
        switch ( action.type ) {

            // Add a block to the entries list by using the corresponding 
            // entry ID as the object key
            case 'SET_BLOCK_BY_ENTRY_ID':
                return {
                    ...state,
                    blocks: {
                        ...state.blocks,
                        [ action.entryId ]: action.block,
                    },
                };

            // Update the bool value for if we've finished initial load
            case 'SET_HAS_FINISHED_INITIAL_LOAD':
                return {
                    ...state,
                    hasFinishedInitialLoad: action.hasFinishedInitialLoad
                };

            // Update the bool value for if we've hit max pages
            case 'SET_HAS_HIT_MAX_PAGES':
                return {
                    ...state,
                    hasHitMaxPages: action.hasHitMaxPages
                };

            // Set the entire array of key events
            case 'SET_KEY_EVENTS':
                return {
                    ...state,
                    keyEvents: action.keyEvents
                };

            // Add a key event and make sure we sort by date
            case 'ADD_KEY_EVENT':
                let updatedKeyEvents = state.keyEvents;

                // Look if the entry already exists
                let existingKey = false;
                updatedKeyEvents.forEach((entry, key) => {
                    if (entry.id === action.entry.id) {
                        existingKey = key;
                    }
                });

                if (existingKey === false) {

                    // Entry doesn't exist; add it
                    updatedKeyEvents.push(action.entry);
                } else {

                    // Entry exists already; update it
                    updatedKeyEvents[ existingKey ] = action.entry;
                }

                return {
                    ...state,
                    keyEvents: updatedKeyEvents.sort((a,b) => {
                        return new Date(b.date) - new Date(a.date);
                    })
                };

            // Remove key event by entry ID by filtering the array
            case 'REMOVE_KEY_EVENT_BY_ENTRY_ID':
                return {
                    ...state,
                    keyEvents: state.keyEvents.filter((entry) => {
                        return entry.id !== action.entryId;
                    })
                };
        }

        return state;
    },
    actions,
    selectors: {

        // Get a block by post object ID
        getBlockByEntryId( state, entryId ) {
            const { blocks } = state;
            const block = blocks[ entryId ];

            return block;
        },

        // Get all blocks in the editor as a post ID => block map
        getBlocks( state ) {
            const { blocks } = state;

            return blocks;
        },

        // Get all key events
        getKeyEvents( state ) {
            const { keyEvents } = state;

            return keyEvents;
        },

        // Flag for if we've finished initial load
        hasFinishedInitialLoad( state ) {
            const { hasFinishedInitialLoad } = state;

            return hasFinishedInitialLoad;
        },

        // Flag for if we've hit max pages
        hasHitMaxPages( state ) {
            const { hasHitMaxPages } = state;

            return hasHitMaxPages;
        }
    },

} );

register( store );
