<?php

namespace DMG;

/**
 * Liveblog Entry Post Type
 * 
 * This is a hidden post type used to store each entry in a dmg_liveblog container post type
 */
class Liveblog_Entry_Post_Type {

	/**
	 * The Post Type name
	 */
	const NAME = 'dmg_liveblog_entry';

	/**
	 * Load the post type
	 *
	 * @return void
	 */
	public static function init(){
		add_action( 'init', [ __CLASS__, 'register' ] );
		add_action( 'init', [ __CLASS__, 'register_block' ] );
		add_action( 'rest_api_init', [ __CLASS__, 'register_meta' ] );
		add_action( 'enqueue_block_editor_assets', [ __CLASS__, 'enqueue_editor_scripts' ] );
		add_filter( 'wp_insert_post_data', [ __CLASS__, 'add_parent_id' ], 10, 2 );
		add_action( 'wp_after_insert_post', [ __CLASS__, 'purge_edge_cache' ], 10, 2 );
	}

	/**
	 * Register the post type
	 *
	 * @return void
	 */
	public static function register(){
		register_post_type(self::NAME,
			[
				'labels'       => [
					'name'          => 'Liveblog entries',
					'singular_name' => 'Liveblog entry',
				],
				'public'       => false,
				'show_in_rest' => true,
				'supports'     => [ 'title', 'author', 'editor' ],
			]
		);
	}

	/**
	 * Register Gutenberg block
	 *
	 * @return void
	 */
	public static function register_block() {
		register_block_type(
			plugin_dir_path( dirname( __FILE__ ) ) . 'gutenberg/src/liveblog-entry/block.json',
			[
				'editor_script'   => 'dmg-liveblog-entry-block',
				'render_callback' => [ __CLASS__, 'render_block_callback' ],
			]
		);
	}

	/**
	 * Render callback method for the block
	 * This is what appears in the content.rendered property of /wp-json/wp/v2/liveblog/$liveblog_id/entries
	 *
	 * @param array  $attributes The blocks attributes
	 * @param string $content    Data returned from InnerBlocks.Content
	 * @param array  $block      Block information such as context.
	 * @return string The rendered block markup.
	 */
	public static function render_block_callback( $attributes, $content, $block ) {

		// get the author assigned to the entry with fallback to the creator of the entry
		$author_id    = ! empty( $attributes['authorOverride'] ) && $attributes['authorOverride'] > 0 ? (int) $attributes['authorOverride'] : (int) $attributes['author'];
		$display_name = ( ! empty( $author_id ) ) ? get_the_author_meta( 'display_name', $author_id ) : '';
		$hide_byline  = ( ! empty( $attributes['entryId'] ) && (bool) get_post_meta( $attributes['entryId'], 'dmg_liveblog_hide_byline', true ) ) ? true : false;

		// co-authors plugin support
		global $coauthors_plus;
		if ( is_a( $coauthors_plus, 'CoAuthors_Plus' ) && method_exists( $coauthors_plus, 'get_coauthor_by' ) ) {
			$coauthor     = $coauthors_plus->get_coauthor_by( 'id', $author_id );
			$display_name = ! empty( $coauthor->display_name ) ? $coauthor->display_name : $display_name;
		}

		// Generate a DateTime object to represent the entry's date in the site's local timezone
		$datetime = new \DateTime( $attributes['date'] );
		$datetime->setTimezone( new \DateTimeZone( wp_timezone_string() ) );

		// Get the post object
		$post = get_post( $attributes['entryId'] );

		/**
		 * Filter the title before output
		 *
		 * @param string  $title_display The existing title
		 * @param WP_Post $entry         The entry post object
		 */
		$title_display = apply_filters( 'dmg_liveblog_entry_title_display', $attributes['title'], $post );

		/**
		 * Filter the author name before output
		 *
		 * @param string  $author_display The content to display for the author (defaults to the display nama)
		 * @param int     $author_id      The author ID
		 * @param WP_Post $entry          The entry post object
		 */
		$author_display = apply_filters( 'dmg_liveblog_entry_author_display', $display_name, $author_id, $post );

		/**
		 * Filter the content before output
		 *
		 * @param string  $author_display The content to display for the author (defaults to the display nama)
		 * @param WP_Post $entry          The entry post object
		 */
		$content_display = apply_filters( 'dmg_liveblog_entry_content', $content, $post );

		ob_start();
		?>
		<div class="dmg-liveblog__entry">
			<?php if ( ! empty( $attributes['date'] ) ) : ?>
				<time class="dmg-liveblog__entry-date" datetime="<?php echo esc_attr( str_replace( ' ', 'T', $post->post_date_gmt ) ) . 'Z'; ?>"><?php echo esc_html( $datetime->format( Liveblog_Post_Type::get_datetime_format() ) ); ?></time>
			<?php endif; ?>
			<?php if ( ! empty( $title_display ) ) : ?>
				<div class="dmg-liveblog__entry-title"><h4><?php echo wp_kses_post( $title_display ); ?></h4></div>
			<?php endif; ?>
			<?php if ( ! $hide_byline ) : ?>
				<div class="dmg-liveblog__entry-author"><?php if ( ! empty( $author_display ) ) { echo wp_kses_post( $author_display ); } ?></div>
			<?php endif; ?>
			<div class="dmg-liveblog__entry-content">
				<?php echo $content_display; ?>
			</div>
		</div>
		<?php
		// WordPress doesn't support core FB embeds out of the box, so VIP relies
		// on Jetpack to render them. On single post templates, the Jetpack enqueues a
		// script and inlines some JSON in the footer. We will try to approximate
		// that approach here when returning an entry via the API.
		if ( (bool) preg_match( '/class="fb-(post|video)"/', $content_display ) ) {
			global $wp_scripts;
			$wp_scripts->do_item( 'jetpack-facebook-embed' );
		}

		// remove unnecessary whitespace
		$entry = str_replace( [ "\n", "\t" ], '', ob_get_clean() );
		$entry = apply_filters( 'dmg_liveblog_entry_template', $entry, $attributes, $content_display );

		return $entry;
	}

	/**
	 * Register custom post meta
	 */
	public static function register_meta() {

		// Register the "Key Event" field
		register_rest_field(
			self::NAME,
			'dmg_liveblog_key_event',
			[
				'get_callback' => function() {
					return (bool) get_post_meta( get_the_ID(), 'dmg_liveblog_key_event', true );
				},
				'update_callback' => function( $value, $object, $field_name ) {
					return update_post_meta( $object->ID, 'dmg_liveblog_key_event', $value );
				},
			]
		);

		// Register the "Hide byline" field
		register_rest_field(
			self::NAME,
			'dmg_liveblog_hide_byline',
			[
				'get_callback' => function() {
					return (bool) get_post_meta( get_the_ID(), 'dmg_liveblog_hide_byline', true );
				},
				'update_callback' => function( $value, $object, $field_name ) {
					return update_post_meta( $object->ID, 'dmg_liveblog_hide_byline', $value );
				},
			]
		);

		// Register the "Author override" meta field
		register_rest_field( self::NAME, 'dmg_liveblog_author_override', [
				'get_callback'    => function() { return (int) get_post_meta( get_the_ID(), 'dmg_liveblog_author_override', true ); },
				'update_callback' => function( $value, $object, $field_name ) { return update_post_meta( $object->ID, 'dmg_liveblog_author_override', (int) $value ); },
			]
		);
	}

	/**
	  * Enqueue block assets in the editor
	  *
	  * @return void
	  */
	public static function enqueue_editor_scripts() {

		// Always enqueue styles because they can appear both in the Liveblog post type and in the Post post type when embedding a Liveblog
		wp_enqueue_style(
			'dmg-liveblog-editor',
			plugin_dir_url( dirname( __FILE__ ) ) . 'gutenberg/build/dmg-liveblog-editor.css',
		);

		// Only enqueue the rest of the assets on liveblog post type
		if ( Liveblog_Post_Type::NAME !== get_post_type() ) {
			return;
		}

		// Enqueue block assets
		$block_asset_file = include( plugin_dir_path( dirname( __FILE__ ) ) . 'gutenberg/build/dmg-liveblog-entry-block.asset.php' );

		wp_enqueue_script(
			'dmg-liveblog-entry-block',
			plugin_dir_url( dirname( __FILE__ ) ) . 'gutenberg/build/dmg-liveblog-entry-block.js',
			$block_asset_file['dependencies'],
			$block_asset_file['version']
		);

		// Enqueue editor assets
		$editor_asset_file = include( plugin_dir_path( dirname( __FILE__ ) ) . 'gutenberg/build/dmg-liveblog-editor.asset.php' );

		wp_enqueue_script(
			'dmg-liveblog-editor',
			plugin_dir_url( dirname( __FILE__ ) ) . 'gutenberg/build/dmg-liveblog-editor.js',
			$editor_asset_file['dependencies'],
    		$editor_asset_file['version']
		);
	}

	/**
	 * Make sure a parent ID is added before inserting an entry into the database
	 *
	 * Core POST requests to update the post don't allow for a parent ID
	 *
	 * @param array $data                An array of slashed, sanitized, and processed post data.
	 * @param array $postarr             An array of sanitized (and slashed) but otherwise unmodified post data.
	 * @param array $unsanitized_postarr An array of slashed yet *unsanitized* and unprocessed post data as
	 *                                   originally passed to wp_insert_post().
	 * @param bool  $update              Whether this is an existing post being updated.
	 */
	public static function add_parent_id( $data, $postarr ) {

		// Scope to entry post type
		if ( isset( $data['post_type'] ) && self::NAME === $data['post_type'] ) {

			// Get the parent passed from the payload and set it in post data
			$request_body = file_get_contents( 'php://input' );
			$request_json = json_decode( $request_body );
			if ( ! empty( $request_json->parent ) ) {
				$data['post_parent'] = (int) $request_json->parent;
			}
		}

		return $data;
	}

	/**
	 * Purge the edge cache
	 *
	 * @param int          $post_id     Post ID.
	 * @param WP_Post      $post        Post object.
	 */
	public static function purge_edge_cache( $post_id, $post ) {

		// Check only for published entries with a parent ID
		if ( function_exists( 'wpcom_vip_purge_edge_cache_for_url' )
			&& is_a( $post, 'WP_Post' )
			&& self::NAME === $post->post_type
			&& 'publish' === $post->post_status
			&& ! empty( $post->post_parent )
		) {
			// Purge the edge cache for the liveblog's /entries/ endpoint
			wpcom_vip_purge_edge_cache_for_url( home_url( "/wp-json/wp/v2/liveblog/{$post->post_parent}/entries/" ) );
		}
	}
}