<?php

namespace DMG;

/**
 * Liveblog container Post Type
 * This post type is the main object editors interact with and create dmg_liveblog_entry children
 */
class Liveblog_Post_Type {

	/**
	 * The Post Type/Block name
	 */
	const NAME = 'dmg_liveblog';

	/**
	 * The capability name
	 *
	 * @var string
	 */
	const CAPABILITY = 'manage_dmg_liveblog';

	/**
	 * Meta key for embedded liveblog ID
	 *
	 * @var string
	 */
	const META_KEY_EMBEDDED_ID = 'dmg_liveblog_embedded_id';

	/**
	 * Meta key for liveblog parent post ID
	 *
	 * @var string
	 */
	const META_KEY_PARENT_ID = 'dmg_liveblog_parent_id';

	/**
	 * Load the post type
	 *
	 * @return void
	 */
	public static function init() {
		add_action( 'init', [ __CLASS__, 'register'] );
		add_filter( 'enter_title_here', [ __CLASS__, 'filter_enter_title' ] );

		// handle multi-users simultaneous editing
		add_action( 'load-post.php', [ __CLASS__, 'remove_edit_lock' ] );
		add_action( 'load-post-new.php', [ __CLASS__, 'remove_edit_lock' ] );

		// Only allow LiveBlog Entries within the post_content of a liveblog.
		add_action( 'init', [ __CLASS__, 'register_block' ] );
		add_action( 'enqueue_block_editor_assets', [ __CLASS__, 'print_settings_var' ] );
		add_action( 'enqueue_block_editor_assets', [ __CLASS__, 'enqueue_editor_scripts' ] );

		// Short-circuit content save
		add_filter( 'wp_insert_post_data', [ __CLASS__, 'add_parent_id' ], 10, 2 );

		// Register custom post_meta
		add_action( 'add_meta_boxes', [ __CLASS__, 'register_custom_meta' ] );
		add_action( 'save_post', [ __CLASS__, 'save_custom_meta' ], 10, 2 );
		add_action( 'rest_api_init', [ __CLASS__, 'register_api_meta' ] );

		// Disable autosaving/backup notices
		add_filter( 'block_editor_settings_all', function( $settings ) {
			$screen = get_current_screen();
			if ( self::NAME === $screen->post_type ) {
				$settings['autosaveInterval'] = 99999999;
				$settings['autosave'] = false;
			}
			return $settings;
		});

		// Always remove any autosave data from browser storage. @TODO there may be a nicer way to do this, but I didn't find it.
		add_action('admin_head', function(){
			$screen = get_current_screen();
			if ( self::NAME === $screen->post_type ) {
				?>
				<script>window.sessionStorage.removeItem('wp-autosave-block-editor-post-<?php echo get_the_ID();?>');</script>
				<?php
			}
		});

		// Save the liveblog ID in meta
		add_action( 'wp_after_insert_post', array( __CLASS__, 'save_liveblog_id_in_meta' ), 10, 2 );
	}

	/**
	 * Register the post type
	 *
	 * @return void
	 */
	public static function register() {
		$labels = [
			'name'               => 'Liveblog',
			'singular_name'      => 'Liveblog',
			'menu_name'          => 'Liveblogs',
			'add_new'            => 'Add New',
			'add_new_item'       => 'Add New Liveblog',
			'edit'               => 'Edit',
			'edit_item'          => 'Edit Liveblog',
			'new_item'           => 'New Liveblog',
			'view'               => 'View Liveblogs',
			'view_item'          => 'View Liveblog',
			'search_items'       => 'Search Liveblogs',
			'not_found'          => 'No Liveblog Found',
			'not_found_in_trash' => 'No Liveblog Found in Trash',
			'parent'             => 'Parent Liveblog',
		];

		register_post_type( self::NAME,
			[
				'has_archive'  => true,
				'hierarchical' => false,
				'labels'       => $labels,
				'menu_icon'    => 'dashicons-excerpt-view',
				'public'       => false,
				'query_var'    => true,
				'show_ui'      => true,
				'show_in_menu' => true,
				'show_in_rest' => true,
				'rest_base'    => 'liveblog',
				'supports'     => [ 'title', 'editor', 'thumbnail', 'excerpt' ],
				'map_meta_cap' => true,
				'capabilities' => [
					'read'                   => 'read',
					'create_posts'           => self::CAPABILITY,
					'edit_posts'             => self::CAPABILITY,
					'delete_posts'           => self::CAPABILITY,
					'publish_posts'          => self::CAPABILITY,
					'edit_published_posts'   => self::CAPABILITY,
					'delete_published_posts' => self::CAPABILITY,
					'edit_others_posts'      => self::CAPABILITY,
					'delete_others_posts'    => self::CAPABILITY,
					'read_private_posts'     => self::CAPABILITY,
					'edit_private_posts'     => self::CAPABILITY,
					'delete_private_posts'   => self::CAPABILITY,
				]
			]
		);
	}

	/**
	 * Filter "Add title" for a nicer UX in the Block editor
	 *
	 * @param string $placeholder The current placeholder
	 * @return string The filtered placeholder
	 */
	public static function filter_enter_title( $placeholder ){
		$screen = get_current_screen();

		if  ( self::NAME === $screen->post_type ) {
			$placeholder = 'Enter liveblog title';
		}

		return $placeholder;
	}

	/**
	 * Disable the post lock check and heartbeat
	 *
	 * @return void
	 */
	public static function remove_edit_lock() {
		if ( self::NAME === get_current_screen()->post_type ) {
			add_filter( 'wp_check_post_lock_window', '__return_zero' );
			wp_deregister_script( 'heartbeat' );
		}
	}

	/**
	 * Register Gutenberg block
	 *
	 * @return void
	 */
	public static function register_block() {
		register_block_type(
			plugin_dir_path( dirname( __FILE__ ) ) . 'gutenberg/src/liveblog/block.json',
			[
				'editor_script'   => 'dmg-liveblog-block',
				'render_callback' => [ __CLASS__, 'render_block_callback' ],
			]
		);
	}

	/**
	 * Print settings var to the page
	 */
	public static function print_settings_var() {

		// Print the liveblog settings to all pages with the block editor
		add_action(
			'admin_print_scripts',
			fn() => printf(
				"<script>var dmgLiveblogSettings = %s;</script>",
				wp_json_encode( self::get_settings() )
			)
		);
	}

	/**
	  * Enqueue block assets in the editor
	  *
	  * @return void
	  */
	public static function enqueue_editor_scripts() {

		// Don't enqueue in the editor
		if ( self::NAME === get_post_type() ) {
			return;
		}

		$asset_file = include( plugin_dir_path( dirname( __FILE__ ) ) . 'gutenberg/build/dmg-liveblog-block.asset.php' );

		wp_enqueue_script(
			'dmg-liveblog-block',
			plugin_dir_url( dirname( __FILE__ ) ) . 'gutenberg/build/dmg-liveblog-block.js',
			$asset_file['dependencies'],
			$asset_file['version']
		);

		wp_enqueue_style(
			'dmg-liveblog-block',
			plugin_dir_url( dirname( __FILE__ ) ) . 'gutenberg/build/dmg-liveblog-block.css',
		);
	}

	/**
	 * Render block callback
	 *
	 * @param array  $attributes The blocks attributes
	 * @param string $content    Data returned from InnerBlocks.Content
	 * @param array  $block      Block information such as context.
	 * @return string The rendered block markup.
	 */
	public static function render_block_callback( $attributes, $content, $block ) {

		$html = '<div id="dmg-liveblog-root"></div>';

		// Don't inline assets on REST reqests; save for later in the request
		if ( ! defined( 'REST_REQUEST' ) || ! REST_REQUEST ) {
			$html .= self::get_inline_assets( $attributes['id'], get_the_ID() );
		}

		return $html;
	}

	/**
	 * Get inline assets
	 *
	 * @param int $liveblog_id The liveblog ID
	 * @param int $post_id     The ID of the post containing the liveblog
	 * @return string A string of inline assets
	 */
	public static function get_inline_assets( $liveblog_id, $post_id ) {

		// Get the app script URL and default styles
		$app_manifest = json_decode( file_get_contents( dirname( dirname( __FILE__ ) ) . '/react-app/build/asset-manifest.json' ) );
		$script_url   = plugin_dir_url( dirname( __FILE__ ) ) . 'react-app/build' . $app_manifest->files->{'main.js'};
		$styles_url   = plugin_dir_url( dirname( __FILE__ ) ) . 'react-app/build' . $app_manifest->files->{'main.css'};

		// Pull out styles for inlining
		$styles = explode( "\n", file_get_contents( $styles_url ) );
		$inline_styles = ( ! empty( $styles[0] ) ) ? $styles[0] : '';

		// Return the root element along with the script
		return sprintf(
			'<style>%s</style><script>var DMGLiveblog = %s</script><script defer src="%s"></script>',
			$inline_styles,
			wp_json_encode( self::get_liveblog_configuration( $liveblog_id, $post_id ) ),
			esc_url( $script_url )
		);
	}

	/**
	 * Get the Liveblog configuration object
	 *
	 * @param int $liveblog_id The liveblog ID
	 * @param int $post_id     The post ID the liveblog is embedded on
	 * @return array The liveblog configuration
	 */
	public static function get_liveblog_configuration( $liveblog_id, $post_id ) {

		/**
		 * Permit filtering before returning configuration
		 *
		 * @param int $liveblog_id The liveblog ID
		 * @param int $post_id     The post ID
		 */
		return apply_filters(
			'dmg_liveblog_configuration',
			array(
				'apiBase'           => home_url( '/wp-json/' ),
				'liveblogID'        => $liveblog_id,
				'embedParentURL'    => get_permalink( $post_id ),// The permalink URL of the WP Post that this liveblog is rendered on
				'showHeadline'      => true, // show/hide the Liveblog post.title
				'showOrderSelector' => true, // render the dropdown to select to order by date ASC/DESC
				'emitSchema'        => true, // render the Schema.org LiveBlogPosting json+ld
				'keyEvents'         => array(
					'active'       => true, // whether or not to render the Key Events
					'count'        => 100, // How many Key Events to show
					'container'    => null, // Could be an array with one of the following properties, which are expected to be valid JS selectors:
						//'selector'     => '#dmg-liveblog-key-events', // Specify a selector to render key events within
						//'insertBefore' => null, // Specify a selector that key events should render before
						//'insertAfter'  => null, // Specify a selector that key events should render after
						//'appendTo'     => null, // Specify a selector that key events should be appended to
				),
				'entries'           => array(
					'count'      => 10, // this is the per_page number to show
				),
				'polling'           => array(
					'active'       => true,
					'frequency'    => 10, // (poll for new entries every X seconds)
					'timestampEnd' => strtotime( get_post_meta( $liveblog_id, 'dmg_liveblog_end_date', true ) ),
				),
				'shareTools'        => array(
					'active'    => true, // render share tools per Entry
					'platforms' => array( // list of share tools to supports
						'facebook',
						'twitter',
						'whatsapp',
						'email',
						'link'
					),
				),
			),
			$liveblog_id,
			$post_id
		);
	}

	/**
	 * Generate the LiveBlogPosting schema
	 * @see https://schema.org/LiveBlogPosting
	 *
	 * @param int $post_id The ID of the liveblog
	 * @return string the schema.org LiveBlogPosting JSON+LD
	 */
	public static function generate_schema( $post_id ) {
		$embed_parents = self::get_post_embed_parents( $post_id );
		$schema = [
			'@context'			=> 'https://schema.org',
			'@type'				=> 'LiveBlogPosting',
			'url' 				=> get_permalink( $embed_parents[0] ),
			'mainEntityOfPage'	=> get_permalink( $embed_parents[0] ),
			'datePublished'		=> get_the_time( 'c', $post_id ),
			'dateModified'		=> get_the_modified_time( 'c', $post_id ),
			'coverageStartTime'	=> get_post_meta( $post_id, 'dmg_liveblog_start_date', true ),
			'coverageEndTime'	=> get_post_meta( $post_id, 'dmg_liveblog_end_date', true ),
			'headline' 			=> get_the_title( $post_id ),
			'description'		=> get_the_excerpt( $post_id ),
			'about' 			=> get_the_excerpt( $post_id ),
			'image'				=> get_the_post_thumbnail_url( $post_id, 'full' ),
			'liveBlogUpdate'	=> self::get_schema_updates( $post_id ),
		];

		return apply_filters( 'dmg_liveblog_schema', $schema, $post_id );
	}

	/**
	 * Get liveBlogUpdate objects for a given liveblog
	 *
	 * @param int $liveblog_id The liveblog ID
	 * @return array The liveblog update object
	 */
	public static function get_schema_updates( $liveblog_id ) {

		$updates = array();

		// Query latest liveblog posts
		$query = new \WP_Query(
			array(
				'post_parent'    => (int) $liveblog_id,
				'post_type'      => Liveblog_Entry_Post_Type::NAME,
				'posts_per_page' => 100 // Limit to latest 100
			)
		);

		// Convert to Schema.org update
		if ( ! empty( $query->posts ) ) {
			$updates = array_filter(
				array_map(
					array( __CLASS__, 'convert_entry_to_schema_update' ),
					$query->posts
				)
			);
		}

		return $updates;
	}

	/**
	 * Convert a post object to a Schema.org liveblog update
	 *
	 * @param WP_Post $post The post object
	 * @return array The update object
	 */
	public static function convert_entry_to_schema_update( $post ) {

		// Require a valid parent post
		$embed_parents = self::get_post_embed_parents( $post->post_parent );
		if ( empty ( $embed_parents[0] ) ) {
			return;
		}

		// Determine permalinks
		$parent_url = get_permalink( $embed_parents[0] );
		$post_url   = $parent_url . '#entry-' . $post->ID;

		// Vars to assist in generating timestamps
		$datetime  = new \DateTime();
		$timestamp = mysql2date( 'G', $post->post_date_gmt );

		// Grab the post content
		$html = array();
		$blocks = parse_blocks( $post->post_content );
		if ( ! empty( $blocks[0]['innerBlocks'] ) ) {
			foreach ( $blocks[0]['innerBlocks'] as $block ) {
				if ( ! empty( $block['innerHTML'] ) ) {
					$html[] = $block['innerHTML'];
				}
			}
		}

		return array(
			'@type'            => 'BlogPosting',
			'mainEntityOfPage' => $parent_url,
			'url'              => $post_url,
			'datePublished'    => $datetime->setTimeStamp( $timestamp )->format( \DateTime::ISO8601 ),
			'dateModified'     => $datetime->setTimeStamp( $timestamp )->format( \DateTime::ISO8601 ),
			'articleBody'      => wp_strip_all_tags( implode( ' ', $html ) ),
			'headline'         => substr( $post->post_title, 0, 110 ), // Truncate to 110 characters
			'author'           => array(
				'name' => get_the_author_meta( 'display_name', $post->post_author ),
			),
		);
	}

	/**
	 * Save content as a blank string because it is inconsequential; all content can
	 * be pulled in from the child entry objects
	 *
	 * @param array $data                An array of slashed, sanitized, and processed post data.
	 * @param array $postarr             An array of sanitized (and slashed) but otherwise unmodified post data.
	 * @param array $unsanitized_postarr An array of slashed yet *unsanitized* and unprocessed post data as
	 *                                   originally passed to wp_insert_post().
	 * @param bool  $update              Whether this is an existing post being updated.
	 */
	public static function add_parent_id( $data, $postarr ) {

		// Scope to liveblog post type
		if ( isset( $data['post_type'] ) && self::NAME === $data['post_type'] ) {
			$data['post_content'] = '';
		}

		return $data;
	}

	/**
	 * Search for posts that contain this liveblog
	 *
	 * @param int $liveblog_id The ID of the liveblog to search
	 * @return array The list of post IDs that contain this liveblog
	 */
	public static function get_post_embed_parents( $liveblog_id ) {
		return get_post_meta( $liveblog_id, self::META_KEY_PARENT_ID );
	}

	/**
	 * Register custom metaboxes
	 *
	 * @return void
	 */
	public static function register_custom_meta() {
		add_meta_box( 'dmg_liveblog_dates', 'Coverage Dates', [ __CLASS__, 'register_date_meta' ], 'dmg_liveblog', 'side', 'high');
	}

	/**
	 * Output post meta fields for the liveblog start/end dates
	 *
	 * @return void
	 */
	public static function register_date_meta() {
		$date_start = get_post_meta( get_the_ID(), 'dmg_liveblog_start_date', true );
		$date_end = get_post_meta( get_the_ID(), 'dmg_liveblog_end_date', true );

		$date_start = ( empty( $date_start ) ) ? gmdate( 'Y-m-d\TH:i', time() ) : gmdate( 'Y-m-d\TH:i', strtotime( $date_start ) );
		$date_end = ( empty( $date_end ) ) ? gmdate( 'Y-m-d\TH:i', strtotime( 'tomorrow' ) ) : gmdate( 'Y-m-d\TH:i', strtotime( $date_end ) );

		?>
		<fieldset class="dmg_liveblog_date_meta">
			<label for="dmg_liveblog_start_date">Start
				<input name="dmg_liveblog_start_date" type="datetime-local" value="<?php echo $date_start; ?>">
			</label>
			<label for="dmg_liveblog_end_date"> End
				<input name="dmg_liveblog_end_date" type="datetime-local" value="<?php echo $date_end; ?>">
			</label>
		</fieldset>
		<?php
	}

	/**
	 * Save custom post meta
	 *
	 * @param int     $post_id The post ID
	 * @param WP_Post $post    The post object
	 * @return void
	 */
	public static function save_custom_meta( $post_id, $post ) {
		
		if ( defined('DOING_AUTOSAVE') && DOING_AUTOSAVE ) {
			return;
		}

		if ( $post->post_type !== self::NAME ) {
			return;
		}

		if ( !current_user_can( self::CAPABILITY ) ){
			return;
		}
		
		if ( isset( $_POST ) ) {
			$start_date = ( !empty( $_POST[ 'dmg_liveblog_start_date' ] ) ) ? $_POST[ 'dmg_liveblog_start_date' ] : gmdate( 'c', time() );
			$end_date = ( !empty( $_POST[ 'dmg_liveblog_end_date' ] ) ) ? $_POST[ 'dmg_liveblog_end_date' ] : gmdate( 'c', strtotime( 'tomorrow' ) );
			
			// Always store the dates in 'c' ISO 8601 format
			update_post_meta( $post->ID, 'dmg_liveblog_start_date', gmdate('c', strtotime( $start_date ) ) );
			update_post_meta( $post->ID, 'dmg_liveblog_end_date', gmdate('c', strtotime( $end_date ) ) );
		}
	}

	 /**
	  * Register meta for API
	  *
	  * @return void
	  */
	public static function register_api_meta() {

		register_rest_field(
			self::NAME,
			'dmg_liveblog_start_date',
			[
				'get_callback' => function() {
					return get_post_meta( get_the_ID(), 'dmg_liveblog_start_date', true );
				},
				'update_callback' => function( $value, $object, $field_name ) {
					return update_post_meta( $object->ID, 'dmg_liveblog_start_date', gmdate('c', strtotime( $value ) ) );
				},
			]
		);

		register_rest_field(
			self::NAME,
			'dmg_liveblog_end_date',
			[
				'get_callback' => function() {
					return get_post_meta( get_the_ID(), 'dmg_liveblog_end_date', true );
				},
				'update_callback' => function( $value, $object, $field_name ) {
					return update_post_meta( $object->ID, 'dmg_liveblog_end_date', gmdate('c', strtotime( $value ) ) );
				},
			]
		);

	}

	/**
	 * Get liveblog settings
	 *
	 * @return array The liveblog settings
	 */
	public static function get_settings() {
		$settings = array(
			'blockWhitelist' => self::get_block_whitelist(),
			'datetimeFormat' => self::get_datetime_format(),
		);

		return apply_filters( 'dmg_liveblog_settings', $settings );
	}

	/**
	 * Get the datetime format
	 *
	 * @return string The datetime format
	 */
	public static function get_datetime_format() {

		// Get site date/time settings with default fallback
		$date_format = get_option( 'date_format', 'F j, Y' );
		$time_format = get_option( 'time_format', 'g:i a' );
		
		// Provide a filter to customize
		return apply_filters( 'dmg_liveblog_datetime_format', "{$date_format} {$time_format}" );
	}

	/**
	 * Get block whitelist
	 *
	 * @return array List of whitelisted blocks
	 */
	public static function get_block_whitelist() {

		// Get an array of registered blocks; whitelist DMG properties and don't permit liveblog blocks
		$registered_blocks = array_values(
			array_filter(
				array_keys( \WP_Block_Type_Registry::get_instance()->get_all_registered() ),
				fn( $block ) => (
					str_starts_with( $block, 'dmg' )
					|| str_starts_with( $block, 'inews' )
					|| str_starts_with( $block, 'metro' )
				) && (
					! in_array( $block, array( 'dmg/liveblog', 'dmg/liveblog-entry' ), true )
				)
			)
		);

		return apply_filters( 'dmg_liveblog_entry_block_whitelist', $registered_blocks );
	}

	/**
	 * Save the ID of the liveblog in post content as meta
	 *
	 * @param int $post_id The post ID
	 * @param \WP_Post $post The post object being saved
	 */
	public static function save_liveblog_id_in_meta( $post_id, $post ) {

		// Autosave/revision and post status check
		if ( wp_is_post_autosave( $post_id )
			|| wp_is_post_revision( $post_id )
			|| 'publish' !== $post->post_status
		) {
			return;
		}

		// Get the existing embedded ID
		$existing_liveblog_id = get_post_meta( $post_id, self::META_KEY_EMBEDDED_ID, true );

		$blocks = parse_blocks( $post->post_content );

		// Find a liveblog block and update meta accordingly
		$found_block = false;
		foreach ( $blocks as $block ) {
			if ( isset( $block['blockName'] )
				&& 'dmg/liveblog' === $block['blockName']
				&& ! empty( $block['attrs']['id'] )
			) {

				$liveblog_id = (int) $block['attrs']['id'];

				// Found the existing liveblog; break
				if ( $liveblog_id === $existing_liveblog_id ) {
					break;
				}

				// Save embedded ID in parent post
				update_post_meta( $post_id, self::META_KEY_EMBEDDED_ID, $liveblog_id );

				// Liveblog changed values
				if ( $existing_liveblog_id !== $liveblog_id ) {
					delete_post_meta( $existing_liveblog_id, self::META_KEY_PARENT_ID, $post_id );
				}

				// Save parent post in liveblog meta
				add_post_meta( $liveblog_id, self::META_KEY_PARENT_ID, $post_id );

				$found_block = true;

				break;
			}
		}

		// Liveblog was removed; delete meta
		if ( ! $found_block && ! empty( $existing_liveblog_id ) ) {
			delete_post_meta( $post_id, self::META_KEY_EMBEDDED_ID );
			delete_post_meta( $existing_liveblog_id, self::META_KEY_PARENT_ID, $post_id );
		}

	}

}