import { useEffect } from 'react';
import { useSelector, useDispatch } from 'react-redux';

/**
 * Internal dependencies
 */
import './App.css';

import Entry from './components/Entry'
import KeyEvents from './components/KeyEvents'
import LoadMore from './components/LoadMore'
import OrderSelector from './components/OrderSelector'
import Queue from './components/Queue'

import { setInitialConfiguration, getConfiguration } from './store/slices/Configuration'
import { fetchLiveblog, getLiveblog } from './store/slices/Liveblog'
import {
  fetchEntries,
  poll,
  fetchAndSeek,
  getEntries,
  reachedMaxPages
} from './store/slices/Entries'

import { ReactComponent as Icons } from './assets/icons.svg'

import store from './store';

const App = () => {

  const dispatch = useDispatch()

  // Get the initial liveblog state using the local configuration
  useEffect(() => {
    if (window?.DMGLiveblog) {
      dispatch(setInitialConfiguration(window.DMGLiveblog));
      dispatch(fetchLiveblog());

      (async () => {

        // Do an initial fetch of entries
        await dispatch(fetchEntries()).then(() => {

          // Set a short timeout to let fetched entries render
          setTimeout(() => {

            // Do we have a hash route?
            if (!window?.location?.hash) {
              return
            }

            // Match on the hash route
            const matches = window.location.hash.match(/#entry-(\d+)/)

            if (matches?.[1]) {

              const entryId = parseInt(matches[1])

              // Look for the entry container
              const entry = document.querySelector(`[data-entry="${entryId}"]`)

              if (entry) {

                // Found the entry; scroll into view
                entry.scrollIntoView()
              } else {

                // Couldn't find it; load in pages until we find it
                document.querySelector('.dmg-liveblog__load-more').scrollIntoView(false)
                dispatch(fetchAndSeek(entryId))
              }
            }
          }, 100)
        })
      })()

      if (!window?.DMGLiveblog?.polling) {
        return;
      }

      const pollingConfig = window.DMGLiveblog.polling;

      // Set up polling
      if (pollingConfig?.active) {

        const currentTime = new Date().getTime();

        // Only start polling if we have not reached the end of coverage
        if (pollingConfig?.timestampEnd && currentTime < pollingConfig.timestampEnd * 1000) {

          // Determine interval
          let pollingInterval = pollingConfig?.frequency ?? 10
          pollingInterval *= 1000 // Convert to milliseconds

          const startPolling = setInterval(() => {

            // Suspend polling when we're not viewing in descending order
            if (store.getState().entries.order !== 'desc') {
              return;
            }

            const pollCurrentTime = new Date().getTime();

            if (pollCurrentTime > pollingConfig.timestampEnd * 1000) {

              // Coverage has ended since we started polling, clear the interval to stop;
              clearInterval(startPolling)
            } else {

              // Continue polling
              dispatch(poll())
            }
          }, pollingInterval)
        }
      }
      
    }
  }, [dispatch]);

  // Get configuration, liveblog and entries objects
  const configuration = useSelector(getConfiguration);
  const liveblog = useSelector(getLiveblog);
  const entries = useSelector(getEntries);

  const hasReachedMaxPages = useSelector(reachedMaxPages);

  return (
    <div className="dmg-liveblog">
      <header className="dmg-liveblog__header">
        {configuration?.showHeadline &&
          liveblog?.title?.rendered &&
          <h3
            dangerouslySetInnerHTML={{
              __html: liveblog.title.rendered
            }}
          />
        }
      </header>
      {configuration?.keyEvents?.active &&
        !configuration?.keyEvents?.container &&
        <KeyEvents />
      }
      {configuration?.showOrderSelector &&
        <OrderSelector />
      }
      <Queue />
      {entries.length > 0 &&
        <div className="dmg-liveblog__entries">
          {entries.map((entry) => <Entry post={entry} key={entry.id} />)}
        </div>
      }
      {!hasReachedMaxPages &&
        <LoadMore />
      }
      <Icons />
    </div>
  );
}

export default App;
