import { createAsyncThunk, createSlice } from '@reduxjs/toolkit'

/**
 * Internal dependencies
 */
import { WP } from '../../utils/wpapi';
import { triggerEvent } from '../../utils/events';

/**
 * Build an entry URL
 *
 * @param {string}   parentURL    The parent post URL
 * @param {number}   entryID      The entry ID
 */
export const getEntryURL = (parentURL, entryID) => {
  return `${parentURL}#entry-${entryID}`
}

/**
 * Create a thunk for fetching entries
 */
export const fetchEntries = createAsyncThunk(
  'entries/fetch',
  async (_: void, thunkAPI) => {
    thunkAPI.dispatch(setIsLoading(true))
    const state = thunkAPI.getState();
    const response = await WP
      .liveblog()
      .id(state.configuration.liveblogID)
      .entries()
      .page(state.entries.page)
      .perPage(state.entries.perPage)
      .order(state.entries.order)
      .param('before', state.entries.timestamp)
      .then((data) => data)
      .catch((err) => err);

    // With the context of the parent post, build unique entry URLs for sharing
    if (!response.error) {
      response.forEach((entry) => {
        if (entry.id) {
          entry.link = getEntryURL(state.configuration.embedParentURL, entry.id)
        }
      })
    }

    return response.error ? [] : response;
  }
)

/**
 * Create a thunk for seeking a given entry ID
 */
export const fetchAndSeek = createAsyncThunk(
  'entries/fetchAndSeek',
  async (entryId, thunkAPI) => {
    thunkAPI.dispatch(setIsLoading(true))
    const state = thunkAPI.getState();

    let page = state.entries.page
    let fetchedEntries = []
    let fetchedEntryIds = []

    if (!state.entries.ids.includes(entryId)) {

      // Loop until we've found the entry we're seeking
      while (!fetchedEntryIds.includes(entryId)) {

        page++

        // Query next page
        let response = await WP
          .liveblog()
          .id(state.configuration.liveblogID)
          .entries()
          .page(page)
          .perPage(state.entries.perPage)
          .order(state.entries.order)
          .param('before', state.entries.timestamp)
          .then((data) => data)
          .catch((err) => err);

        if (!response.error) {

          // Build IDs array and generate share URLs
          response.forEach((entry) => {
            if (entry.id) {
              fetchedEntryIds.push(entry.id)
              entry.link = getEntryURL(state.configuration.embedParentURL, entry.id)
            }
          })

          // Build entries array
          fetchedEntries = fetchedEntries.concat(response)
        }
      }
    }

    return {
      fetchedEntries,
      entryId,
      page
    }
  }
)

/**
 * Create a thunk for polling
 */
export const poll = createAsyncThunk(
  'entries/poll',
  async (_: void, thunkAPI) => {
    const state = thunkAPI.getState();

    // Pluck out the latest post, either from the queue or the rendered entries
    let latestPost = false
    if (state?.entries?.queue[0]) {
      latestPost = state.entries.queue[0]
    } else if (state?.entries?.posts[0]) {
      latestPost = state.entries.posts[0]
    }

    // Get timestamp from GMT date
    const dateString = latestPost.date_gmt.replace( 'T', ' ' ) + ' UTC';
    const afterDate = new Date(dateString)
    const afterTimestamp = Math.floor(afterDate.getTime() / 10000) * 10

    const response = await WP
      .liveblog()
      .id(state.configuration.liveblogID)
      .entries()
      .param('after', afterTimestamp)
      .then((data) => data)
      .catch((err) => err);

    // With the context of the parent post, build unique entry URLs for sharing
    if (!response.error) {
      response.forEach((entry) => {
        if (entry.id) {
          entry.link = getEntryURL(state.configuration.embedParentURL, entry.id)
        }
      })
    }

    return response.error ? [] : response;
  }
)

/**
 * Function for determining if top of the liveblog is in view
 */
export const isTopOfLiveblogInView = () => {
  let inView = false
  const queueButton = document.querySelector('.dmg-liveblog__queue');
  if (queueButton) {

    // Check if the queue button is in the viewport
    const rect = document.querySelector('.dmg-liveblog__queue').getBoundingClientRect();
    if ( rect.top >= 0 &&
        rect.left >= 0 &&
        rect.bottom <= (window.innerHeight || document.documentElement.clientHeight) &&
        rect.right <= (window.innerWidth || document.documentElement.clientWidth)
    ) {
      inView = true
    }
  }

  return inView
}

/**
 * Collect scripts from a payload and load them asynchronously
 *
 * @{param}
 */
const asyncLoadScripts = (payload) => {

  let scriptsToLoad = []

  // Check if the scripts have been loaded
  payload.forEach((entry) => {
    const matches = entry.content.rendered.matchAll(/<script.*?src=("|')(.*?)("|')/g)
    for (const match of matches) {
      if (!scriptsToLoad.includes(match[2])) {
        scriptsToLoad.push(match[2])
      }
    }
  })

  // Load scripts
  if (scriptsToLoad.length > 0) {

    scriptsToLoad.forEach((script) => {

      // Not crazy about this, but if we are receiving a Jetpack Facebook embed,
      // the script expects a variable that is a JSON with WordPress.com's
      // app ID. Declare that here before we print the script to the page.
      if (script.includes('facebook') && !window.jpfbembed) {
        window.jpfbembed = {"appid":"249643311490"};
      }

      // Search for the script on the page
      const existingScript = document.querySelector(`script[src^="${script}"]`)

      // Remove existing script and affix a timestamp to the src to force reload
      if (existingScript) {
        existingScript.remove()
        script += `?${Date.now()}`
      }

      // Create a new script tag
      const newScript = document.createElement('script');
      newScript.src = script;
      newScript.async = false;
      document.head.appendChild(newScript);
    })
  }
}

/**
 * Create a slice
 */
export const entries = createSlice({
  name: 'entries',
  initialState: {
    posts: [],
    queue: [],
    ids: [],
    page: 1,
    perPage: 10,
    order: 'desc',
    maxPages: 1,
    loading: false,
    timestamp: Math.floor(Date.now() / 10000) * 10 // Divide by 10,000 to convert to seconds, plus round to nearest 10th
  },
  reducers: {

    // Increment the page
    incrementPage: (state) => {
      state.page += 1
    },

    // Reset the page value
    resetPage: (state) => {
      state.page = 1
    },

    // Change the order; reset the posts array
    setOrder: (state, action) => {
      if (['asc','desc'].includes(action.payload.order.toLowerCase())) {
        state.order = action.payload.order
        state.posts = []
        state.ids = []
      }
    },

    // Load queued posts
    loadQueue: (state, action) => {
      asyncLoadScripts(state.queue)
      state.posts = state.queue.concat(state.posts)
      state.queue = []
    },

    // Set the loading state
    setIsLoading: (state, action) => {
      state.loading = action.payload
    }
  },
  extraReducers: (builder) => {
    
    // When entries are successfully fetched, add them to the entries array
    builder.addCase(fetchEntries.fulfilled, (state, action) => {

      asyncLoadScripts(action.payload)

      // Add new posts
      state.posts = state.posts.concat(action.payload)

      // Add new IDs
      state.ids = state.ids.concat(action.payload.map((post) => {
        return post.id
      }))

      // Set total pages
      if (action?.payload?._paging?.totalPages) {
        state.maxPages = action.payload._paging.totalPages;
      }

      // Toggle loading state off
      state.loading = false

      triggerEvent('dmgLiveblog:fetchEntries', action.payload)
    })

    // When polling and we retrieve an entries, add them
    builder.addCase(poll.fulfilled, (state, action) => {

      // Update the timestamp
      if (action.payload.length > 0
        && action.payload[0]?.id
        && !state.ids.includes(action.payload[0].id)
      ) {
        state.timestamp = Math.floor(Date.now() / 10000) * 10 
      }

      // Loop through each returned item in the payload
      action.payload.forEach((entry) => {

        // Check against our array of IDs
        if (!state.ids.includes(entry.id)) {
          state.queue.unshift(entry)
          state.ids.unshift(entry.id)
        }
      })

      // Sort the queue by date
      state.queue.sort((a, b) => {
        const dateA = new Date(a.date)
        const dateB = new Date(b.date)
        return dateB - dateA
      })

      // If the top of the liveblog is in view, automatically load the entries in
      if (isTopOfLiveblogInView()) {
        asyncLoadScripts(state.queue)
        state.posts = state.queue.concat(state.posts)
        state.queue = []
      }

      triggerEvent('dmgLiveblog:poll', action.payload)
    })

    // When polling and we retrieve an entries, add them
    builder.addCase(fetchAndSeek.fulfilled, (state, action) => {

      asyncLoadScripts(action.payload.fetchedEntries)

      // Add new posts
      state.posts = state.posts.concat(action.payload.fetchedEntries)

      // Add new IDs
      state.ids = state.ids.concat(action.payload.fetchedEntries.map((post) => {
        return post.id
      }))

      // Toggle loading state off
      state.loading = false

      // Update the page counter
      state.page = action.payload.page

      // Scroll selected event into view
      setTimeout(() => document.querySelector(`[data-entry="${action.payload.entryId}"]`).scrollIntoView(), 100)

      triggerEvent('dmgLiveblog:fetchAndSeek', action.payload)
    })
  }
})

export const { incrementPage, resetPage, setOrder, loadQueue, setIsLoading } = entries.actions

// Entries selectors
export const getEntries = (state) => state.entries.posts
export const reachedMaxPages = (state) => state.entries.page >= state.entries.maxPages
export const getQueue = (state) => state.entries.queue
export const isLoading = (state) => state.entries.loading

export default entries.reducer
